# TODO: Standardize / unify this code with ag.save()
import os, pickle, tempfile, logging, boto3

from .. import s3_utils

logger = logging.getLogger(__name__)


def save(path, object, format=None, verbose=True):
    pickle_fn = lambda o, buffer: pickle.dump(o, buffer, protocol=4)
    save_with_fn(path, object, pickle_fn, format=format, verbose=verbose)


def save_with_fn(path, object, pickle_fn, format=None, verbose=True):
    if verbose:
        logger.log(15, 'Saving '+str(path))
    if s3_utils.is_s3_url(path):
        format = 's3'
    if format == 's3':
        save_s3(path, object, pickle_fn, verbose=verbose)
    else:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as fout:
            pickle_fn(object, fout)


def save_s3(path: str, obj, pickle_fn, verbose=True):
    if verbose:
        logger.info(f'save object to {path}')
    with tempfile.TemporaryFile() as f:
        pickle_fn(obj, f)
        f.flush()
        f.seek(0)

        bucket, key = s3_utils.s3_path_to_bucket_prefix(path)
        s3_client = boto3.client('s3')
        try:
            config = boto3.s3.transfer.TransferConfig()   # enable multipart uploading for files larger than 8MB
            response = s3_client.upload_fileobj(f, bucket, key, Config=config)
        except:
            logger.exception('Failed to save object to s3')
            raise



from typing import Tuple, List, Optional
import mxnet as mx
from mxnet.ndarray import NDArray
import numpy as np
import logging

from autogluon.searcher.bayesopt.datatypes.tuning_job_state import \
    TuningJobState
from autogluon.searcher.bayesopt.tuning_algorithms.base_classes import \
    SurrogateModel
from autogluon.searcher.bayesopt.gpmxnet.constants import DATA_TYPE
from autogluon.searcher.bayesopt.autogluon.debug_log import DebugLogPrinter

logger = logging.getLogger(__name__)


class SurrogateModelMXNet(SurrogateModel):
    """
    MXNet based model

    """
    def __init__(
            self, state: TuningJobState, active_metric: str, random_seed: int,
            ctx_nd=None, dtype_nd=None,
            debug_log: Optional[DebugLogPrinter] = None):
        if ctx_nd is None:
            ctx_nd = mx.cpu()
        if dtype_nd is None:
            dtype_nd = DATA_TYPE
        super(SurrogateModelMXNet, self).__init__(
            state, active_metric, random_seed)
        assert dtype_nd == np.float32 or dtype_nd == np.float64
        self.ctx_nd = ctx_nd
        self.dtype_nd = dtype_nd
        self._current_best = None
        self._debug_log = debug_log

    ######################################
    # shared functions for all sub-classes

    def predict(self, X: np.ndarray) -> List[Tuple[np.ndarray, np.ndarray]]:
        """return prediction of X using predict_nd"""
        dtype_np, _ = self._get_dtypes(X)
        if X.ndim == 1:
            X = X[None, :]

        X_nd = self.convert_np_to_nd(X)
        predictions_nd = self.predict_nd(X_nd)

        predictions_list = [
            (m_nd.asnumpy().astype(dtype_np, copy=False),
             s_nd.asnumpy().astype(dtype_np, copy=False))
            for m_nd, s_nd in predictions_nd
        ]
        return predictions_list

    def context_for_nd(self) -> mx.Context:
        return self.ctx_nd

    def dtype_for_nd(self):
        return self.dtype_nd

    def _get_dtypes(self, x: np.ndarray):
        dtype_np = x.dtype
        if dtype_np == np.int32 or dtype_np == np.int64:
            dtype_np = np.float64
        assert dtype_np == np.float32 or dtype_np == np.float64
        dtype_nd = self.dtype_nd
        return dtype_np, dtype_nd

    def _current_best_filter_candidates(self, candidates):
        """
        Can be used by subclasses to specialize current_best.

        """
        return candidates

    def current_best(self) -> np.ndarray:
        if self._current_best is None:
            def convert(c):
                x = self.state.hp_ranges.to_ndarray(c)
                return self.convert_np_to_nd(x).reshape((1, -1))

            candidates = [
                x.candidate for x in self.state.candidate_evaluations] + \
                         self.state.pending_candidates
            if self._debug_log is not None:
                deb_msg = "[GPMXNetModel.current_best -- RECOMPUTING]\n"
                deb_msg += ("- len(candidates) = {}".format(len(candidates)))
                logger.info(deb_msg)

            if len(candidates) == 0:
                # this can happen if no evaluation is present for the current task
                # in some sense the best so far is plus infinity, any new result would be an improvement
                # In this special setting optimizing the EI is equivalent to optimizing the mean
                # and any current_best high enough will approximately give the same results
                assert len(self.state.candidate_evaluations) > 0  # other tasks will have evaluations
                # take the worst and add a few times standard dev
                values = [x.metrics[self.active_metric] for x in self.state.candidate_evaluations]
                current_worst = max(values)
                std = np.std(values)
                self._current_best = np.array([current_worst + 10 * std])
            else:
                candidates = self._current_best_filter_candidates(candidates)
                X_all = mx.nd.concat(*map(convert, candidates), dim=0)
                pred_mean = _compute_mean_across_samples(self.predict_nd(X_all))
                self._current_best = mx.nd.min(pred_mean, axis=0).asnumpy()

            logger.info(f"Current best is {self._current_best}")
        return self._current_best


def _compute_mean_across_samples(predictions_list: List[Tuple[NDArray, NDArray]]) -> NDArray:
    pred_means_list = [pred_mean for pred_mean, _ in predictions_list]
    pred_means = mx.nd.stack(*pred_means_list, axis=0)
    return mx.nd.mean(pred_means, axis=0)



import numpy as np
import mxnet as mx
from mxnet import autograd
from mxnet.ndarray import NDArray
from abc import ABC, abstractmethod
from typing import Tuple, Optional

from autogluon.searcher.bayesopt.tuning_algorithms.base_classes import \
    SurrogateModel, AcquisitionFunction
from autogluon.searcher.bayesopt.models.nphead_acqfunc import \
    _reshape_predictions


class MXNetHeadAcquisitionFunction(AcquisitionFunction, ABC):
    """
    Base class for acquisition functions whose head are implemented in MXNet.
    Here:

        f(x, model) = h(mean, std, model.current_best())

    where h(.) is the head, and mean, std are predictive mean and stddev.

    """
    def __init__(self, model: SurrogateModel):
        super(MXNetHeadAcquisitionFunction, self).__init__(model)

    def compute_acq(self, x: np.ndarray,
                    model: Optional[SurrogateModel] = None) -> np.ndarray:
        if model is None:
            model = self.model
        predictions_list = model.predict_nd(model.convert_np_to_nd(x))
        fvals_list = []
        for mean, std in predictions_list:
            if self._head_needs_current_best():
                current_best = model.convert_np_to_nd(
                    model.current_best()).reshape((1, -1))
            else:
                current_best = None
            # Are we in batch mode? If so, mean is a matrix, whose number of columns
            # is identical to the number of samples used for fantasizing. In this
            # case, we compute the criterion values separate for every sample, and
            # average them
            do_avg = (mean.ndim == 2 and mean.shape[1] > 1)
            if do_avg:
                assert mean.shape[1] == current_best.size, \
                    "mean.shape[1] = {}, current_best.size = {} (must be the same)".format(
                        mean.shape[1], current_best.size)
                std = std.reshape((-1, 1))
            fvals = self._compute_head(mean, std, current_best)
            if do_avg:
                fvals = mx.nd.mean(fvals, axis=1)
            fvals_list.append(fvals.asnumpy().astype(x.dtype, copy=False))
        return np.mean(fvals_list, axis=0)

    def compute_acq_with_gradients(
            self, x: np.ndarray,
            model: Optional[SurrogateModel] = None) -> \
            Tuple[np.ndarray, np.ndarray]:
        if model is None:
            model = self.model
        dtype_np = x.dtype
        if x.ndim == 1:
            x = x[None, :]
        num_data = x.shape[0]

        # Loop over cases (rows of x), we need the gradients for each case
        # separately
        f_acqu = np.empty((num_data, 1), dtype=dtype_np)
        df_acqu = np.empty_like(x)
        # The current best
        if self._head_needs_current_best():
            current_best_nd = model.convert_np_to_nd(
                model.current_best()).reshape((-1,))
        else:
            current_best_nd = None

        for row in range(num_data):
            x_nd = model.convert_np_to_nd(x[row, None])
            # Record for gradient computation
            x_nd.attach_grad()
            with autograd.record():
                m_nd, s_nd = _reshape_predictions(model.predict_nd(x_nd))
                fval = mx.nd.mean(self._compute_head(
                    m_nd, s_nd, current_best_nd))
            f_acqu[row] = fval.asscalar()
            fval.backward()
            df_acqu[row] = x_nd.grad.asnumpy().astype(
                dtype_np, copy=False)
        return f_acqu, df_acqu

    @abstractmethod
    def _head_needs_current_best(self) -> bool:
        """
        :return: Is the current_best argument in _compute_head needed?
        """
        pass

    @abstractmethod
    def _compute_head(self, mean: NDArray, std: NDArray,
                      current_best: NDArray) -> NDArray:
        """
        All input and return arguments are of type mx.nd.NDArray.

        :param mean: Predictive means
        :param std: Predictive stddevs
        :param current_best: Encumbent
        :return: hvals
        """
        pass


class LCBAcquisitionFunction(MXNetHeadAcquisitionFunction):
    """
    Lower confidence bound (LCB) acquisition function:

        h(mean, std) = mean - kappa * std

    """
    def __init__(self, model: SurrogateModel, kappa: float):
        super(LCBAcquisitionFunction, self).__init__(model)
        assert kappa > 0, 'kappa must be positive'
        self.kappa = kappa

    def _head_needs_current_best(self) -> bool:
        return False

    def _compute_head(self, mean: NDArray, std: NDArray,
                      current_best: NDArray) -> NDArray:
        return mean - self.kappa * std



import math
import random
import numbers
import warnings
import numpy as np
import mxnet as mx
from PIL import Image, ImageEnhance
import collections

__all__ = ['Compose', 'RandomResizedCrop', 'RandomHorizontalFlip', 'ColorJitter',
           'Resize', 'CenterCrop', 'ToTensor', 'RandomCrop', 'ToNDArray', 'ToPIL']

class Compose(object):
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

class ToPIL(object):
    """Convert image from ndarray format to PIL
    """
    def __call__(self, img):
        x = Image.fromarray(img.asnumpy())
        return x

class ToNDArray(object):
    def __call__(self, img):
        x = mx.nd.array(np.array(img), mx.cpu(0))
        return x

class ToTensor(object):
    def __call__(self, img):
        x = mx.nd.array(np.array(img), mx.cpu(0))
        return x.swapaxes(0, 2).swapaxes(1, 2)

class RandomResizedCrop(object):
    """Crop the given PIL Image to random size and aspect ratio.
    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.
    Args:
        size: expected output size of each edge
        scale: range of size of the origin size cropped
        ratio: range of aspect ratio of the origin aspect ratio cropped
        interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.), interpolation=Image.BILINEAR):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (size, size)
        if (scale[0] > scale[1]) or (ratio[0] > ratio[1]):
            warnings.warn("range should be of kind (min, max)")

        self.interpolation = interpolation
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        width, height = img.size
        area = height * width

        for attempt in range(10):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = random.randint(0, height - h)
                j = random.randint(0, width - w)
                return i, j, h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if (in_ratio < min(ratio)):
            w = width
            h = int(round(w / min(ratio)))
        elif (in_ratio > max(ratio)):
            h = height
            w = int(round(h * max(ratio)))
        else:  # whole image
            w = width
            h = height
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped and resized.
        Returns:
            PIL Image: Randomly cropped and resized image.
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        return self.resized_crop(img, i, j, h, w, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string

    @staticmethod
    def resized_crop(img, top, left, height, width, size, interpolation=Image.BILINEAR):
        img = crop(img, top, left, height, width)
        img = resize(img, size, interpolation)
        return img
    
class RandomCrop(object):
    """Crop the given PIL Image at a random location.

    Parameters
    ----------
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is None, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively. If a sequence of length 2 is provided, it is used to
            pad left/right, top/bottom borders, respectively.
        pad_if_needed (boolean): It will pad the image if smaller than the
            desired size to avoid raising an exception. Since cropping is done
            after padding, the padding seems to be done at a random offset.
        fill: Pixel fill value for constant fill. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.
            This value is only used when the padding_mode is constant
        padding_mode: Type of padding. Should be: constant, edge, reflect or symmetric. Default is constant.

             - constant: pads with a constant value, this value is specified with fill

             - edge: pads with the last value on the edge of the image

             - reflect: pads with reflection of image (without repeating the last value on the edge)

                padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                will result in [3, 2, 1, 2, 3, 4, 3, 2]

             - symmetric: pads with reflection of image (repeating the last value on the edge)

                padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                will result in [2, 1, 1, 2, 3, 4, 4, 3]

    """
    def __init__(self, size, padding=None, pad_if_needed=False, fill=0, padding_mode='constant'):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Parameters
        ----------
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns
        -------
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h = _get_image_size(img)
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped.

        Returns:
            PIL Image: Cropped image.
        """
        if self.padding is not None:
            img = pad(img, self.padding, self.fill, self.padding_mode)

        # pad the width if needed
        if self.pad_if_needed and img.size[0] < self.size[1]:
            img = pad(img, (self.size[1] - img.size[0], 0), self.fill, self.padding_mode)
        # pad the height if needed
        if self.pad_if_needed and img.size[1] < self.size[0]:
            img = pad(img, (0, self.size[0] - img.size[1]), self.fill, self.padding_mode)

        i, j, h, w = self.get_params(img, self.size)

        return crop(img, i, j, h, w)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)


class RandomHorizontalFlip(object):
    """Horizontally flip the given PIL Image randomly with a given probability.
    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be flipped.
        Returns:
            PIL Image: Randomly flipped image.
        """
        if random.random() < self.p:
            return img.transpose(Image.FLIP_LEFT_RIGHT)
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)

class Lambda(object):
    """Apply a user-defined lambda as a transform.
    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        assert callable(lambd), repr(type(lambd).__name__) + " object is not callable"
        self.lambd = lambd

    def __call__(self, img):
        return self.lambd(img)

    def __repr__(self):
        return self.__class__.__name__ + '()'

class ColorJitter(object):
    """Randomly change the brightness, contrast and saturation of an image.
    Args:
        brightness (float or tuple of float (min, max)): How much to jitter brightness.
            brightness_factor is chosen uniformly from [max(0, 1 - brightness), 1 + brightness]
            or the given [min, max]. Should be non negative numbers.
        contrast (float or tuple of float (min, max)): How much to jitter contrast.
            contrast_factor is chosen uniformly from [max(0, 1 - contrast), 1 + contrast]
            or the given [min, max]. Should be non negative numbers.
        saturation (float or tuple of float (min, max)): How much to jitter saturation.
            saturation_factor is chosen uniformly from [max(0, 1 - saturation), 1 + saturation]
            or the given [min, max]. Should be non negative numbers.
        hue (float or tuple of float (min, max)): How much to jitter hue.
            hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
            Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    """
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        self.brightness = self._check_input(brightness, 'brightness')
        self.contrast = self._check_input(contrast, 'contrast')
        self.saturation = self._check_input(saturation, 'saturation')
        self.hue = self._check_input(hue, 'hue', center=0, bound=(-0.5, 0.5),
                                     clip_first_on_zero=False)

    def _check_input(self, value, name, center=1, bound=(0, float('inf')), clip_first_on_zero=True):
        if isinstance(value, numbers.Number):
            if value < 0:
                raise ValueError("If {} is a single number, it must be non negative.".format(name))
            value = [center - value, center + value]
            if clip_first_on_zero:
                value[0] = max(value[0], 0)
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            if not bound[0] <= value[0] <= value[1] <= bound[1]:
                raise ValueError("{} values should be between {}".format(name, bound))
        else:
            raise TypeError("{} should be a single number or a list/tuple with lenght 2.".format(name))

        # if value is 0 or (1., 1.) for brightness/contrast/saturation
        # or (0., 0.) for hue, do nothing
        if value[0] == value[1] == center:
            value = None
        return value

    @staticmethod
    def get_params(brightness, contrast, saturation, hue):
        """Get a randomized transform to be applied on image.
        Arguments are same as that of __init__.
        Returns:
            Transform which randomly adjusts brightness, contrast and
            saturation in a random order.
        """
        transforms = []

        if brightness is not None:
            brightness_factor = random.uniform(brightness[0], brightness[1])
            transforms.append(Lambda(lambda img: adjust_brightness(img, brightness_factor)))

        if contrast is not None:
            contrast_factor = random.uniform(contrast[0], contrast[1])
            transforms.append(Lambda(lambda img: adjust_contrast(img, contrast_factor)))

        if saturation is not None:
            saturation_factor = random.uniform(saturation[0], saturation[1])
            transforms.append(Lambda(lambda img: adjust_saturation(img, saturation_factor)))

        if hue is not None:
            hue_factor = random.uniform(hue[0], hue[1])
            transforms.append(Lambda(lambda img: adjust_hue(img, hue_factor)))

        random.shuffle(transforms)
        transform = Compose(transforms)

        return transform

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Input image.
        Returns:
            PIL Image: Color jittered image.
        """
        transform = self.get_params(self.brightness, self.contrast,
                                    self.saturation, self.hue)
        return transform(img)

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += 'brightness={0}'.format(self.brightness)
        format_string += ', contrast={0}'.format(self.contrast)
        format_string += ', saturation={0}'.format(self.saturation)
        format_string += ', hue={0})'.format(self.hue)
        return format_string

class Resize(object):
    """Resize the input PIL Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be scaled.
        Returns:
            PIL Image: Rescaled image.
        """
        return resize(img, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        return self.__class__.__name__ + '(size={0}, interpolation={1})'.format(self.size, interpolate_str)


class CenterCrop(object):
    """Crops the given PIL Image at the center.
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be cropped.
        Returns:
            PIL Image: Cropped image.
        """
        return self.center_crop(img, self.size)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0})'.format(self.size)

    @staticmethod
    def center_crop(img, output_size):
        if isinstance(output_size, numbers.Number):
            output_size = (int(output_size), int(output_size))
        image_width, image_height = img.size
        crop_height, crop_width = output_size
        crop_top = int(round((image_height - crop_height) / 2.))
        crop_left = int(round((image_width - crop_width) / 2.))
        return crop(img, crop_top, crop_left, crop_height, crop_width)

def crop(img, top, left, height, width):
    return img.crop((left, top, left + width, top + height))

def resize(img, size, interpolation=Image.BILINEAR):
    if not (isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)):
        raise TypeError('Got inappropriate size arg: {}'.format(size))

    if isinstance(size, int):
        w, h = img.size
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return img.resize((ow, oh), interpolation)
        else:
            oh = size
            ow = int(size * w / h)
            return img.resize((ow, oh), interpolation)
    else:
        return img.resize(size[::-1], interpolation)

def adjust_brightness(img, brightness_factor):
    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(brightness_factor)
    return img

def adjust_contrast(img, contrast_factor):
    enhancer = ImageEnhance.Contrast(img)
    img = enhancer.enhance(contrast_factor)
    return img

def adjust_saturation(img, saturation_factor):
    enhancer = ImageEnhance.Color(img)
    img = enhancer.enhance(saturation_factor)
    return img


def adjust_hue(img, hue_factor):
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))

    input_mode = img.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return img
    h, s, v = img.convert('HSV').split()
    np_h = np.array(h, dtype=np.uint8)
    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(hue_factor * 255)
    h = Image.fromarray(np_h, 'L')

    img = Image.merge('HSV', (h, s, v)).convert(input_mode)
    return img

def pad(img, padding, fill=0, padding_mode='constant'):
    if not isinstance(padding, (numbers.Number, tuple)):
        raise TypeError('Got inappropriate padding arg')
    if not isinstance(fill, (numbers.Number, str, tuple)):
        raise TypeError('Got inappropriate fill arg')
    if not isinstance(padding_mode, str):
        raise TypeError('Got inappropriate padding_mode arg')

    if isinstance(padding, Sequence) and len(padding) not in [2, 4]:
        raise ValueError("Padding must be an int or a 2, or 4 element tuple, not a " +
                         "{} element tuple".format(len(padding)))

    assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric'], \
        'Padding mode should be either constant, edge, reflect or symmetric'

    if padding_mode == 'constant':
        if img.mode == 'P':
            palette = img.getpalette()
            image = ImageOps.expand(img, border=padding, fill=fill)
            image.putpalette(palette)
            return image
        return ImageOps.expand(img, border=padding, fill=fill)
    else:
        if isinstance(padding, int):
            pad_left = pad_right = pad_top = pad_bottom = padding
        if isinstance(padding, Sequence) and len(padding) == 2:
            pad_left = pad_right = padding[0]
            pad_top = pad_bottom = padding[1]
        if isinstance(padding, Sequence) and len(padding) == 4:
            pad_left = padding[0]
            pad_top = padding[1]
            pad_right = padding[2]
            pad_bottom = padding[3]
        if img.mode == 'P':
            palette = img.getpalette()
            img = np.asarray(img)
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right)), padding_mode)
            img = Image.fromarray(img)
            img.putpalette(palette)
            return img

        img = np.asarray(img)
        # RGB image
        if len(img.shape) == 3:
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right), (0, 0)), padding_mode)
        # Grayscale image
        if len(img.shape) == 2:
            img = np.pad(img, ((pad_top, pad_bottom), (pad_left, pad_right)), padding_mode)

        return Image.fromarray(img)



# Do not change these!
BINARY = 'binary'
MULTICLASS = 'multiclass'
REGRESSION = 'regression'
SOFTCLASS = 'softclass' # classification with soft-target (rather than classes, labels are probabilities of each class).

PROBLEM_TYPES_CLASSIFICATION = [BINARY, MULTICLASS]
PROBLEM_TYPES_REGRESSION = [REGRESSION]
PROBLEM_TYPES = PROBLEM_TYPES_CLASSIFICATION + PROBLEM_TYPES_REGRESSION + [SOFTCLASS]

REFIT_FULL_NAME = 'refit_single_full'  # stack-name used for refit_single_full (aka "compressed") models
REFIT_FULL_SUFFIX = "_FULL"  # suffix appended to model name for refit_single_full (aka "compressed") models

# AG_ARGS variables are key names in model hyperparameters to dictionaries of custom AutoGluon arguments.
AG_ARGS = 'AG_args'  # Contains arguments to control model name, model priority, and the valid configurations which it can be used in.
AG_ARGS_FIT = 'AG_args_fit'  # Contains arguments that impact model training, such as early stopping rounds, #cores, #gpus, max time limit, max memory usage  # TODO

OBJECTIVES_TO_NORMALIZE = ['log_loss', 'pac_score', 'soft_log_loss']  # do not like predicted probabilities = 0



from mxnet import gluon

from autogluon.searcher.bayesopt.gpmxnet.kernel import KernelFunction, \
    Matern52, ExponentialDecayResourcesKernelFunction, \
    ExponentialDecayResourcesMeanFunction
from autogluon.searcher.bayesopt.gpmxnet.warping import WarpedKernel, Warping


def resource_kernel_factory(
        name: str, kernel_x: KernelFunction, mean_x: gluon.HybridBlock,
        max_metric_value: float) -> (KernelFunction, gluon.HybridBlock):
    """
    Given kernel function kernel_x and mean function mean_x over config x,
    create kernel and mean functions over (x, r), where r is the resource
    attribute (nonnegative scalar, usually in [0, 1]).

    :param name: Selects resource kernel type
    :param kernel_x: Kernel function over configs x
    :param mean_x: Mean function over configs x
    :return: res_kernel, res_mean, both over (x, r)

    """
    if name == 'matern52':
        res_kernel = Matern52(dimension=kernel_x.dimension + 1, ARD=True)
        res_mean = mean_x
    elif name == 'matern52-res-warp':
        # Warping on resource dimension (last one)
        dim_x = kernel_x.dimension
        res_warping = Warping(
            dimension=dim_x + 1, index_to_range={dim_x: (0., 1.)})
        res_kernel = WarpedKernel(
            kernel=Matern52(dimension=dim_x + 1, ARD=True),
            warping=res_warping)
        res_mean = mean_x
    else:
        if name == 'exp-decay-sum':
            delta_fixed_value = 0.0
        elif name == 'exp-decay-combined':
            delta_fixed_value = None
        elif name == 'exp-decay-delta1':
            delta_fixed_value = 1.0
        else:
            raise AssertionError("name = '{}' not supported".format(name))
        res_kernel = ExponentialDecayResourcesKernelFunction(
            kernel_x, mean_x, gamma_init=0.5 * max_metric_value,
            delta_fixed_value=delta_fixed_value,
            max_metric_value=max_metric_value)
        res_mean = ExponentialDecayResourcesMeanFunction(
            kernel=res_kernel)

    return res_kernel, res_mean



from typing import Optional
import numpy as np
from scipy.linalg import solve_triangular
import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH

from autogluon.searcher.bayesopt.models.gpmxnet import \
    get_internal_candidate_evaluations
from autogluon.searcher.bayesopt.datatypes.tuning_job_state import \
    TuningJobState
from autogluon.searcher.bayesopt.autogluon.hp_ranges import \
    HyperparameterRanges_CS
from autogluon.searcher.bayesopt.datatypes.common import CandidateEvaluation
from autogluon.searcher.bayesopt.tuning_algorithms.default_algorithm import \
    dictionarize_objective, DEFAULT_METRIC


class ThreeHumpCamel(object):
    @property
    def search_space(self):
        return [{'min': -5.0, 'max': 5.0},
                {'min': -5.0, 'max': 5.0}]

    def evaluate(self, x1, x2):
        return 2 * x1 ** 2 - 1.05 * x1 ** 4 + x1 ** 6 / 6 + x1 * x2 + x2 ** 2


def branin_function(x1, x2):
    return (x2 - (5.1 / (4 * np.pi ** 2)) * x1 ** 2 + (5 / np.pi) * x1 - 6) ** 2 + \
           10 * (1 - 1 / (8 * np.pi)) * np.cos(x1) + 10


class Branin(object):
    @property
    def search_space(self):
        return [{'min': -5.0, 'max': 10.0},
                {'min': 0.0, 'max': 15.0}]

    def evaluate(self, x1, x2):
        return branin_function(x1, x2)


class Ackley(object):
    @property
    def search_space(self):
        const = 32.768
        return [{'min': -const, 'max': const},
                {'min': -const, 'max': const}]

    def evaluate(self, x1, x2):
        a = 20
        b = 0.2
        c = 2 * np.pi
        ssq = (x1 ** 2) + (x2 ** 2)
        scos = np.cos(c * x1) + np.cos(c * x2)
        return -a * np.exp(-b * np.sqrt(0.5 * ssq)) - np.exp(0.5 * scos) + \
               (a + np.exp(1))


def _decode_input(x, lim):
    mn, mx = lim['min'], lim['max']
    return x * (mx - mn) + mn


def evaluate_blackbox(bb_func, inputs: np.ndarray) -> np.ndarray:
    num_dims = inputs.shape[1]
    input_list = []
    for x, lim in zip(np.split(inputs, num_dims, axis=1), bb_func.search_space):
        input_list.append(_decode_input(x, lim))
    return bb_func.evaluate(*input_list)


# NOTE: Inputs will always be in [0, 1] (so come in encoded form). They are
# only scaled to their native ranges (linearly) when evaluations of the
# blackbox are done. This avoids silly errors.
def sample_data(
        bb_cls, num_train: int, num_grid: int) -> dict:
    bb_func = bb_cls()
    ss_limits = bb_func.search_space
    num_dims = len(ss_limits)
    # Sample training inputs
    train_inputs = np.random.uniform(
        low=0.0, high=1.0, size=(num_train, num_dims))
    # Training targets (currently, no noise is added)
    train_targets = evaluate_blackbox(bb_func, train_inputs).reshape((-1,))
    # Inputs for prediction (regular grid)
    grids = [np.linspace(0.0, 1.0, num_grid)] * num_dims
    grids2 = tuple(np.meshgrid(*grids))
    test_inputs = np.hstack([x.reshape(-1, 1) for x in grids2])
    # Also evaluate true function on grid
    true_targets = evaluate_blackbox(bb_func, test_inputs).reshape((-1,))
    data = {
        'ss_limits': ss_limits,
        'train_inputs': train_inputs,
        'train_targets': train_targets,
        'test_inputs': test_inputs,
        'grid_shape': grids2[0].shape,
        'true_targets': true_targets}
    # Make sure that ours and GPy below receive exactly the same inputs
    state = data_to_state(data)
    data_internal = get_internal_candidate_evaluations(
        state, active_metric=DEFAULT_METRIC, normalize_targets=True,
        num_fantasize_samples=20)
    data['state'] = state
    data['train_inputs'] = data_internal.X
    data['train_targets_normalized'] = data_internal.y
    return data


# Recall that inputs in data are encoded, so we have to decode them to their
# native ranges for candidate_evaluations
def data_to_state(data: dict) -> TuningJobState:
    cs = CS.ConfigurationSpace()
    cs_names = ['x{}'.format(i) for i in range(len(data['ss_limits']))]
    cs.add_hyperparameters([
        CSH.UniformFloatHyperparameter(
            name=name, lower=lims['min'], upper=lims['max'])
        for name, lims in zip(cs_names, data['ss_limits'])])
    _evaluations = []
    x_mult = []
    x_add = []
    for lim in data['ss_limits']:
        mn, mx = lim['min'], lim['max']
        x_mult.append(mx - mn)
        x_add.append(mn)
    x_mult = np.array(x_mult)
    x_add = np.array(x_add)
    for x, y in zip(data['train_inputs'], data['train_targets']):
        x_decoded = x * x_mult + x_add
        config_dct = dict(zip(cs_names, x_decoded))
        config = CS.Configuration(cs, values=config_dct)
        _evaluations.append(CandidateEvaluation(
            config, dictionarize_objective(y)))
    return TuningJobState(
        hp_ranges=HyperparameterRanges_CS(cs),
        candidate_evaluations=_evaluations,
        failed_candidates=[],
        pending_evaluations=[])


def assert_equal_candidates(candidates1, candidates2, hp_ranges, decimal=5):
    inputs1 = hp_ranges.to_ndarray_matrix(candidates1)
    inputs2 = hp_ranges.to_ndarray_matrix(candidates2)
    np.testing.assert_almost_equal(inputs1, inputs2, decimal=decimal)


def assert_equal_randomstate(randomstate1, randomstate2):
    assert str(randomstate1.get_state()) == str(randomstate2.get_state())


def compare_gpy_predict_posterior_marginals(
        test_intermediates: dict, noise_variance_gpy: Optional[float] = None):
    """
    Compares all intermediates of cholesky_computations and
    predict_posterior_marginals to using GPy and NumPy.

    Currently, this is restricted:
    - Kernel must be Matern52 with ARD
    - Mean function must be constant 0

    :param test_intermediates: Intermediates computed using our code
    :param noise_variance_gpy: Overrides noise_variance in test_intermediates.
        Use this if jitter was added during the posterior state computation.

    """
    import GPy
    # Create GPy kernel and model
    num_data = test_intermediates['features'].shape[0]
    num_dims = test_intermediates['features'].shape[1]
    lengthscales = [
        1.0 / test_intermediates['inv_bw{}'.format(i)]
        for i in range(num_dims)]
    kernel = GPy.kern.Matern52(
        num_dims,
        variance=test_intermediates['covariance_scale'],
        lengthscale=lengthscales,
        ARD=True)
    if noise_variance_gpy is None:
        noise_variance_gpy = test_intermediates['noise_variance']
    model = GPy.models.GPRegression(
        test_intermediates['features'],
        test_intermediates['targets'].reshape((-1, 1)),
        kernel=kernel, noise_var=noise_variance_gpy)
    # Compare intermediates step by step (cholesky_computations)
    kernel_mat_gpy = kernel.K(test_intermediates['features'], X2=None)
    np.testing.assert_almost_equal(
        test_intermediates['kernel_mat'], kernel_mat_gpy, decimal=5)
    sys_mat_gpy = kernel_mat_gpy + np.diag(np.ones(num_data)) * \
                  noise_variance_gpy
    np.testing.assert_almost_equal(
        test_intermediates['sys_mat'], sys_mat_gpy, decimal=5)
    chol_fact_gpy = np.linalg.cholesky(sys_mat_gpy)
    # Use test_intermediates['sys_mat'] instead:
    #chol_fact_gpy = np.linalg.cholesky(test_intermediates['sys_mat'])
    np.testing.assert_almost_equal(
        test_intermediates['chol_fact'], chol_fact_gpy, decimal=4)
    # Mean function must be constant 0
    centered_y = test_intermediates['targets'].reshape((-1, 1))
    np.testing.assert_almost_equal(
        test_intermediates['centered_y'], centered_y, decimal=9)
    pred_mat_gpy = solve_triangular(chol_fact_gpy, centered_y, lower=True)
    np.testing.assert_almost_equal(
        test_intermediates['pred_mat'], pred_mat_gpy, decimal=3)
    # Compare intermediates step by step (predict_posterior_marginals)
    k_tr_te_gpy = kernel.K(test_intermediates['features'],
                           X2=test_intermediates['test_features'])
    np.testing.assert_almost_equal(
        test_intermediates['k_tr_te'], k_tr_te_gpy, decimal=5)
    linv_k_tr_te_gpy = solve_triangular(chol_fact_gpy, k_tr_te_gpy, lower=True)
    np.testing.assert_almost_equal(
        test_intermediates['linv_k_tr_te'], linv_k_tr_te_gpy, decimal=4)
    pred_means_gpy = np.dot(linv_k_tr_te_gpy.T, pred_mat_gpy)
    np.testing.assert_almost_equal(
        test_intermediates['pred_means'], pred_means_gpy, decimal=4)
    k_tr_diag_gpy = kernel.Kdiag(
        test_intermediates['test_features']).reshape((-1,))
    tvec_gpy = np.sum(np.square(linv_k_tr_te_gpy), axis=0).reshape((-1,))
    pred_vars_gpy = k_tr_diag_gpy - tvec_gpy
    np.testing.assert_almost_equal(
        test_intermediates['pred_vars'], pred_vars_gpy, decimal=4)
    # Also test against GPy predict
    pred_means_gpy2, pred_vars_gpy2 = model.predict(
        test_intermediates['test_features'], include_likelihood=False)
    pred_vars_gpy2 = pred_vars_gpy2.reshape((-1,))
    np.testing.assert_almost_equal(pred_means_gpy, pred_means_gpy2, decimal=3)
    np.testing.assert_almost_equal(pred_vars_gpy, pred_vars_gpy2, decimal=3)



import numpy as np
import ConfigSpace as CS
from typing import List, Tuple

from autogluon.searcher.bayesopt.datatypes.common import Candidate
from autogluon.searcher.bayesopt.datatypes.hp_ranges import HyperparameterRanges


class HyperparameterRanges_CS(HyperparameterRanges):
    def __init__(self, config_space: CS.ConfigurationSpace,
                 name_last_pos: str = None,
                 value_for_last_pos = None):
        """
        If name_last_pos is given, the hyperparameter of that name is assigned
        the final position in the vector returned by to_ndarray. This can be
        used to single out the (time) resource for a GP model, where that
        component has to come last.

        If in this case (name_last_pos given), value_for_last_pos is also given,
        some methods are modified:
        - random_candidate samples a config as normal, but then overwrites the
          name_last_pos component by value_for_last_pos
        - get_ndarray_bounds works as normal, but returns bound (a, a) for
          name_last_pos component, where a is the internal value corresponding
          to value_for_last_pos
        The use case is HPO with a resource attribute. This attribute should be
        fixed when optimizing the acquisition function, but can take different
        values in the evaluation data (coming from all previous searches).

        :param config_space: ConfigurationSpace
        :param name_last_pos: See above. Default: None
        :param value_for_last_pos: See above. Default: None
        """
        self.config_space = config_space
        self.name_last_pos = name_last_pos
        self.value_for_last_pos = value_for_last_pos
        # Supports conversion to ndarray
        numer_src = []
        numer_trg = []
        categ_src = []
        categ_trg = []
        categ_card = []
        trg_pos = 0
        append_at_end = None
        for src_pos, hp in enumerate(config_space.get_hyperparameters()):
            if isinstance(hp, CS.CategoricalHyperparameter):
                card = hp.num_choices
                if hp.name == name_last_pos:
                    assert append_at_end is None
                    append_at_end = (src_pos, card, True)
                else:
                    categ_src.append(src_pos)
                    categ_trg.append(trg_pos)
                    categ_card.append(card)
                    trg_pos += card
            elif isinstance(hp, CS.UniformIntegerHyperparameter) or \
                    isinstance(hp, CS.UniformFloatHyperparameter):
                if hp.name == name_last_pos:
                    assert append_at_end is None
                    append_at_end = (src_pos, 1, False)
                else:
                    numer_src.append(src_pos)
                    numer_trg.append(trg_pos)
                    trg_pos += 1
            else:
                raise NotImplementedError(
                    "We only support hyperparameters of type "
                    "CategoricalHyperparameter, UniformIntegerHyperparameter, "
                    "UniformFloatHyperparameter")
        if append_at_end is not None:
            if append_at_end[2]:
                categ_src.append(append_at_end[0])
                categ_trg.append(trg_pos)
                categ_card.append(append_at_end[1])
            else:
                numer_src.append(append_at_end[0])
                numer_trg.append(trg_pos)
            trg_pos += append_at_end[1]
        self.numer_src = np.array(numer_src, dtype=np.int64)
        self.numer_trg = np.array(numer_trg, dtype=np.int64)
        self.categ_src = np.array(categ_src, dtype=np.int64)
        self.categ_trg = np.array(categ_trg, dtype=np.int64)
        self.categ_card = np.array(categ_card, dtype=np.int64)
        self._ndarray_size = trg_pos

    def to_ndarray(self, cand_tuple: Candidate) -> np.ndarray:
        assert isinstance(cand_tuple, CS.Configuration)
        trgvec = np.zeros((self._ndarray_size,))
        srcvec = cand_tuple.get_array()
        # https://wesmckinney.com/blog/numpy-indexing-peculiarities/
        # take, put much faster than []
        trgvec.put(
            self.numer_trg, srcvec.take(self.numer_src, mode='clip'),
            mode='clip')
        relpos = srcvec.take(self.categ_src, mode='clip').astype(np.int64)
        trgvec.put(self.categ_trg + relpos, [1.], mode='clip')
        return trgvec

    def ndarray_size(self) -> int:
        return self._ndarray_size

    def from_ndarray(self, cand_ndarray: np.ndarray) -> Candidate:
        assert cand_ndarray.size == self._ndarray_size, \
            "Internal vector [{}] must have size {}".format(
                cand_ndarray, self._ndarray_size)
        cand_ndarray = cand_ndarray.reshape((-1,))
        assert cand_ndarray.min() >= 0. and cand_ndarray.max() <= 1., \
            "Internal vector [{}] must have entries in [0, 1]".format(
                cand_ndarray)
        # Deal with categoricals by using argmax
        srcvec = np.zeros(self.__len__(), dtype=cand_ndarray.dtype)
        srcvec.put(
            self.numer_src, cand_ndarray.take(self.numer_trg, mode='clip'),
            mode='clip')
        for srcpos, trgpos, card in zip(
                self.categ_src, self.categ_trg, self.categ_card):
            maxpos = cand_ndarray[trgpos:(trgpos + card)].argmax()
            srcvec[srcpos] = maxpos
        # Rest is dealt with by CS.Configuration
        return CS.Configuration(self.config_space, vector=srcvec)

    def is_attribute_fixed(self):
        return (self.name_last_pos is not None) and \
               (self.value_for_last_pos is not None)

    def _fix_attribute_value(self, name):
        return self.is_attribute_fixed() and name == self.name_last_pos

    def get_ndarray_bounds(self) -> List[Tuple[float, float]]:
        bounds = []
        final_bound = None
        for hp in self.config_space.get_hyperparameters():
            if isinstance(hp, CS.CategoricalHyperparameter):
                if not self._fix_attribute_value(hp.name):
                    bound = [(0., 1.)] * len(hp.choices)
                else:
                    bound = [(0., 0.)] * len(hp.choices)
                    bound[int(self.value_for_last_pos)] = (1., 1.)
            else:
                if not self._fix_attribute_value(hp.name):
                    bound = [(0., 1.)]
                else:
                    val_int = float(hp._inverse_transform(
                        np.array([self.value_for_last_pos])).item())
                    bound = [(val_int, val_int)]
            if hp.name == self.name_last_pos:
                final_bound = bound
            else:
                bounds.extend(bound)
        if final_bound is not None:
            bounds.extend(final_bound)
        return bounds

    # NOTE: Assumes that config argument not used afterwards...
    def _transform_config(self, config: CS.Configuration) -> CS.Configuration:
        values = config.get_dictionary()  # No copy is done here
        values[self.name_last_pos] = self.value_for_last_pos
        return CS.Configuration(self.config_space, values=values)

    def random_candidate(self, random_state) -> Candidate:
        self.config_space.random = random_state  # Not great...
        rnd_config = self.config_space.sample_configuration()
        if self.is_attribute_fixed():
            rnd_config = self._transform_config(rnd_config)
        return rnd_config

    def random_candidates(
            self, random_state, num_configs: int) -> List[Candidate]:
        self.config_space.random = random_state  # Not great...
        rnd_configs = self.config_space.sample_configuration(num_configs)
        if self.is_attribute_fixed():
            rnd_configs = [self._transform_config(x) for x in rnd_configs]
        return rnd_configs

    def __repr__(self) -> str:
        return "{}{}".format(
            self.__class__.__name__, repr(self.config_space))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HyperparameterRanges_CS):
            return self.config_space == other.config_space
        return False

    def __len__(self) -> int:
        return len(self.config_space.get_hyperparameters())

    # This is just for convenience, roughly shadows what is done in
    # HyperparameterRanges_Impl
    @property
    def hp_ranges(self) -> List[CS.hyperparameters.Hyperparameter]:
        return self.config_space.get_hyperparameters()

    def filter_for_last_pos_value(
            self, candidates: List[Candidate]) -> List[Candidate]:
        """
        If is_attribute_fixed, the candidates list is filtered by removing
        entries whose name_last_pos attribute value is different from
        value_for_last_pos. Otherwise, candidates is returned unchanged.

        """
        if self.is_attribute_fixed():
            def filter_pred(x: CS.Configuration) -> bool:
                x_dct = x.get_dictionary()
                return (x_dct[self.name_last_pos] == self.value_for_last_pos)

            candidates = list(filter(filter_pred, candidates))
        return candidates



from typing import Union
import mxnet as mx
import numpy as np

from autogluon.searcher.bayesopt.gpmxnet.constants import \
    NOISE_VARIANCE_LOWER_BOUND, MIN_POSTERIOR_VARIANCE, \
    MIN_CHOLESKY_DIAGONAL_VALUE
from autogluon.searcher.bayesopt.gpmxnet.custom_op import AddJitterOp, \
    AddJitterOpProp


Tensor = Union[mx.nd.NDArray, mx.sym.Symbol]


def mxnet_F(x: Tensor):
    if isinstance(x, mx.nd.NDArray):
        return mx.nd
    else:
        return mx.sym


def mxnet_is_ndarray(F):
    return F.__name__ == 'mxnet.ndarray'


def cholesky_computations(
        F, features, targets, mean, kernel, noise_variance, debug_log=False,
        test_intermediates=None):
    """
    Given input matrix X (features), target matrix Y (targets), mean and kernel
    function, compute posterior state {L, P}, where L is the Cholesky factor
    of
        k(X, X) + sigsq_final * I
    and
        L P = Y - mean(X)
    Here, sigsq_final >= noise_variance is minimal such that the Cholesky
    factorization does not fail.

    :param F: mx.nd or mx.sym
    :param features: Input matrix X
    :param targets: Target matrix Y
    :param mean: Mean function
    :param kernel: Kernel function
    :param noise_variance: Noise variance (may be increased)
    :param debug_log: Debug output during add_jitter CustomOp?
    :param test_intermediates: If given, all intermediates are written into
        this dict (only if F = mx.nd)
    :return: L, P

    """
    kernel_mat = kernel(features, features)
    # Add jitter to noise_variance (if needed) in order to guarantee that
    # Cholesky factorization works
    sys_mat = F.Custom(
        kernel_mat, noise_variance, name="add_jitter", op_type='add_jitter',
        initial_jitter_factor=NOISE_VARIANCE_LOWER_BOUND,
        debug_log='true' if debug_log else 'false')
    chol_fact = F.linalg.potrf(sys_mat)
    centered_y = F.broadcast_sub(
        targets, F.reshape(mean(features), shape=(-1, 1)))
    pred_mat = F.linalg.trsm(chol_fact, centered_y)
    # For testing:
    if test_intermediates is not None and mxnet_is_ndarray(F):
        assert isinstance(test_intermediates, dict)
        test_intermediates.update({
            'features': features.asnumpy(),
            'targets': targets.asnumpy(),
            'noise_variance': noise_variance.asscalar(),
            'kernel_mat': kernel_mat.asnumpy(),
            'sys_mat': sys_mat.asnumpy(),
            'chol_fact': chol_fact.asnumpy(),
            'pred_mat': pred_mat.asnumpy(),
            'centered_y': centered_y.asnumpy()})
        test_intermediates.update(kernel.get_params())
        test_intermediates.update(mean.get_params())
    return chol_fact, pred_mat


def predict_posterior_marginals(
        F, features, mean, kernel, chol_fact, pred_mat, test_features,
        test_intermediates=None):
    """
    Computes posterior means and variances for test_features.
    If pred_mat is a matrix, so will be posterior_means, but not
    posterior_variances. Reflects the fact that for GP regression and fixed
    hyperparameters, the posterior mean depends on the targets y, but the
    posterior covariance does not.

    :param F: mx.sym or mx.nd
    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :return: posterior_means, posterior_variances

    """
    k_tr_te = kernel(features, test_features)
    linv_k_tr_te = F.linalg.trsm(chol_fact, k_tr_te)
    posterior_means = F.broadcast_add(
        F.linalg.gemm2(
            linv_k_tr_te, pred_mat, transpose_a=True, transpose_b=False),
        F.reshape(mean(test_features), shape=(-1, 1)))
    posterior_variances = kernel.diagonal(F, test_features) - F.sum(
        F.square(linv_k_tr_te), axis=0)
    # For testing:
    if test_intermediates is not None and mxnet_is_ndarray(F):
        assert isinstance(test_intermediates, dict)
        test_intermediates.update({
            'k_tr_te': k_tr_te.asnumpy(),
            'linv_k_tr_te': linv_k_tr_te.asnumpy(),
            'test_features': test_features.asnumpy(),
            'pred_means': posterior_means.asnumpy(),
            'pred_vars': F.reshape(F.maximum(
                posterior_variances, MIN_POSTERIOR_VARIANCE),
                shape=(-1,)).asnumpy()})
    return posterior_means, F.reshape(F.maximum(
        posterior_variances, MIN_POSTERIOR_VARIANCE), shape=(-1,))


def sample_posterior_marginals(
        F, features, mean, kernel, chol_fact, pred_mat, test_features,
        num_samples=1):
    """
    Draws num_sample samples from the product of marginals of the posterior
    over input points test_features. If pred_mat is a matrix with m columns,
    the samples returned have shape (n_test, m, num_samples).

    Note: If F = mx.sym, the samples always have shape
    (n_test, m, num_samples), even if m = 1.

    :param F: mx.sym or mx.nd
    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :param num_samples: Number of samples to draw
    :return: Samples, shape (n_test, num_samples) or (n_test, m, num_samples)

    """
    # Shape of post_means is (n_test, m)
    post_means, post_vars = predict_posterior_marginals(
        F, features, mean, kernel, chol_fact, pred_mat, test_features)
    post_means = F.expand_dims(post_means, axis=-1)  # (n_test, m, 1)
    post_stds = F.sqrt(F.reshape(post_vars, shape=(-1, 1, 1)))  # (n_test, 1, 1)
    n01_vecs = [F.random.normal_like(post_means) for _ in range(num_samples)]
    n01_mat = F.concat(*n01_vecs, dim=-1)  # (n_test, m, num_samples)
    samples = F.broadcast_add(F.broadcast_mul(n01_mat, post_stds), post_means)
    # Remove m == 1 dimension. Can be done only for mx.nd
    if mxnet_is_ndarray(F) and samples.shape[1] == 1:
        samples = F.reshape(samples, shape=(0, -1))  # (n_test, num_samples)
    return samples


def sample_posterior_joint(
        F, features, mean, kernel, chol_fact, pred_mat, test_features,
        num_samples=1):
    """
    Draws num_sample samples from joint posterior distribution over inputs
    test_features. This is done by computing mean and covariance matrix of
    this posterior, and using the Cholesky decomposition of the latter. If
    pred_mat is a matrix with m columns, the samples returned have shape
    (n_test, m, num_samples).

    Note: If F = mx.sym, the samples always have shape
    (n_test, m, num_samples), even if m = 1.

    :param F: mx.sym or mx.nd
    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :param num_samples: Number of samples to draw
    :return: Samples, shape (n_test, num_samples) or (n_test, m, num_samples)

    """
    k_tr_te = kernel(features, test_features)
    linv_k_tr_te = F.linalg.trsm(chol_fact, k_tr_te)
    posterior_mean = F.broadcast_add(
        F.linalg.gemm2(
            linv_k_tr_te, pred_mat, transpose_a=True, transpose_b=False),
        F.reshape(mean(test_features), shape=(-1, 1)))
    posterior_cov = kernel(test_features, test_features) - F.linalg.syrk(
        linv_k_tr_te, transpose=True)
    # Compute the Cholesky decomposition of posterior_cov (expensive!)
    # Add some jitter proactively (>= 1e-5)
    ones_like_test = F.ones_like(F.reshape(F.slice_axis(
        F.BlockGrad(test_features), axis=1, begin=0, end=1), shape=(-1,)))
    jitter_init = F.ones_like(
        F.slice_axis(ones_like_test, axis=0, begin=0, end=1)) * (1e-5)
    sys_mat = F.Custom(
        posterior_cov, jitter_init, name="add_jitter",
        op_type='add_jitter', initial_jitter_factor=NOISE_VARIANCE_LOWER_BOUND)
    lfact = F.linalg.potrf(sys_mat)
    # Draw samples
    # posterior_mean.shape = (n_test, m), where m is number of cols of pred_mat
    # Reshape to (n_test, m, 1)
    posterior_mean = F.expand_dims(posterior_mean, axis=-1)
    n01_vecs = [
        F.random.normal_like(posterior_mean) for _ in range(num_samples)]
    # n01_mat.shape = (n_test, m, num_samples) -> (n_test, m * num_samples)
    n01_mat = F.reshape(F.concat(*n01_vecs, dim=-1), shape=(0, -1))
    # Reshape samples back to (n_test, m, num_samples) after trmm
    samples = F.reshape(
        F.linalg.trmm(lfact, n01_mat), shape=(0, -1, num_samples))
    samples = F.broadcast_add(samples, posterior_mean)
    # Remove m == 1 dimension. Can be done only for mx.nd
    if mxnet_is_ndarray(F) and samples.shape[1] == 1:
        samples = F.reshape(samples, shape=(0, -1))  # (n_test, num_samples)
    return samples


def _compute_lvec(F, features, chol_fact, kernel, feature):
    kvec = F.reshape(kernel(features, feature), shape=(-1, 1))
    return F.reshape(F.linalg.trsm(chol_fact, kvec), shape=(1, -1))


def cholesky_update(
        F, features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature, target, lvec=None):
    """
    Incremental update of posterior state (Cholesky factor, prediction
    matrix), given one datapoint (feature, target).

    Note: noise_variance is the initial value, before any jitter may have
    been added to compute chol_fact. Here, we add the minimum amount of
    jitter such that the new diagonal entry of the Cholesky factor is
    >= MIN_CHOLESKY_DIAGONAL_VALUE. This means that if cholesky_update is
    used several times, we in fact add a diagonal (but not spherical)
    jitter matrix.

    :param F:
    :param features: Shape (n, d)
    :param chol_fact: Shape (n, n)
    :param pred_mat: Shape (n, m)
    :param mean:
    :param kernel:
    :param noise_variance:
    :param feature: Shape (1, d)
    :param target: Shape (1, m)
    :param lvec: If given, this is the new column of the Cholesky factor
        except the diagonal entry. If not, this is computed here
    :return: chol_fact_new (n+1, n+1), pred_mat_new (n+1, m)

    """
    if lvec is None:
        lvec = _compute_lvec(F, features, chol_fact, kernel, feature)
    kscal = F.reshape(kernel.diagonal(F, feature), shape=(1,))
    noise_variance = F.reshape(noise_variance, shape=(1,))
    lsqscal = F.maximum(kscal + noise_variance - F.sum(F.square(lvec)),
                        MIN_CHOLESKY_DIAGONAL_VALUE ** 2)
    lscal = F.reshape(F.sqrt(lsqscal), shape=(1, 1))
    mscal = F.reshape(mean(feature), shape=(1, 1))
    pvec = F.broadcast_sub(target, mscal)
    pvec = F.broadcast_div(pvec - F.linalg.gemm2(lvec, pred_mat), lscal)
    pred_mat_new = F.concat(pred_mat, pvec, dim=0)
    tmpmat = F.concat(chol_fact, lvec, dim=0)
    zerovec = F.zeros_like(F.reshape(lvec, shape=(-1, 1)))
    chol_fact_new = F.concat(
        tmpmat, F.concat(zerovec, lscal, dim=0), dim=1)
    return chol_fact_new, pred_mat_new


# Specialized routine, used in GPPosteriorStateIncrementalUpdater.
# The idea is to share the computation of lvec between sampling a new target
# value and incremental Cholesky update
def sample_and_cholesky_update(
        F, features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature):
    # Draw sample target. Also, lvec is reused below
    lvec = _compute_lvec(F, features, chol_fact, kernel, feature)
    pred_mean = F.broadcast_add(
        F.dot(lvec, pred_mat), F.reshape(mean(feature), shape=(1, 1)))
    # Note: We do not add noise_variance to the predictive variance
    pred_std = F.reshape(F.sqrt(F.maximum(
        kernel.diagonal(F, feature) - F.sum(F.square(lvec)),
        MIN_POSTERIOR_VARIANCE)), shape=(1, 1))
    n01mat = F.random.normal_like(pred_mean)
    target = pred_mean + F.broadcast_mul(n01mat, pred_std)
    # Incremental posterior update
    chol_fact_new, pred_mat_new = cholesky_update(
        F, features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature, target, lvec=lvec)
    features_new = F.concat(features, feature, dim=0)
    return chol_fact_new, pred_mat_new, features_new, target


def negative_log_marginal_likelihood(F, chol_fact, pred_mat):
    """
    The marginal likelihood is only computed if pred_mat has a single column
    (not for fantasy sample case).
    """

    if mxnet_is_ndarray(F):
        assert pred_mat.ndim == 1 or pred_mat.shape[1] == 1, \
            "Multiple target vectors are not supported"
    sqnorm_predmat = F.sum(F.square(pred_mat))
    logdet_cholfact = 2.0 * F.sum(F.log(F.abs(F.diag(chol_fact))))
    # pred_mat is a vector of dimension the number of samples
    n_samples = F.sum(F.ones_like(F.BlockGrad(pred_mat)))
    return 0.5 * (sqnorm_predmat + n_samples * np.log(2 * np.pi) + logdet_cholfact)



import multiprocessing as mp
import warnings
from typing import AnyStr
import pandas as pd

from mxnet import gluon
from mxnet.metric import Accuracy, F1, MCC, PearsonCorrelation, CompositeEvalMetric
from ...utils.try_import import try_import_gluonnlp
from ...core import *
from ...utils.dataset import get_split_samplers, SampledDataset

__all__ = ['MRPCTask', 'QQPTask', 'QNLITask', 'RTETask', 'STSBTask', 'CoLATask', 'MNLITask',
           'WNLITask', 'SSTTask', 'AbstractGlueTask', 'get_dataset', 'AbstractCustomTask']

@func()
def get_dataset(name=None, *args, **kwargs):
    """Load a text classification dataset to train AutoGluon models on.
        
        Parameters
        ----------
        name : str
            Name describing which built-in popular text dataset to use (mostly from the GLUE NLP benchmark).
            Options include: 'mrpc', 'qqp', 'qnli', 'rte', 'sts-b', 'cola', 'mnli', 'wnli', 'sst', 'toysst'. 
            Detailed descriptions can be found in the file: `autogluon/task/text_classification/dataset.py`
        **kwargs : some keyword arguments when using custom dataset as below.
        kwargs['label']: str or int Default: last column
            Index or column name of the label/target column.
    """
    path = kwargs.get('filepath', None)
    if path is not None:
        if path.endswith('.tsv') or path.endswith('.csv'):
            return CustomTSVClassificationTask(*args, **kwargs)
        else:
            raise ValueError('tsv or csv format is supported now, please use the according files ending with'
                             '`.tsv` or `.csv`.')
    if name is not None and name.lower() in built_in_tasks:
        return built_in_tasks[name.lower()](*args, **kwargs)
    else:
        raise ValueError('mrpc, qqp, qnli, rte, sts-b, cola, mnli, wnli, sst, toysst are supported now.')

class AbstractGlueTask:
    """Abstract task class for datasets with GLUE format.

    Parameters
    ----------
    class_labels : list of str, or None
        Classification labels of the task.
        Set to None for regression tasks with continuous real values.
    metrics : list of EValMetric
        Evaluation metrics of the task.
    is_pair : bool
        Whether the task deals with sentence pairs or single sentences.
    label_alias : dict
        label alias dict, some different labels in dataset actually means
        the same. e.g.: {'contradictory':'contradiction'} means contradictory
        and contradiction label means the same in dataset, they will get
        the same class id.
    """
    def __init__(self, class_labels, metrics, is_pair, label_alias=None):
        self.class_labels = class_labels
        self.metrics = metrics
        self.is_pair = is_pair
        self.label_alias = label_alias

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for the task.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments.

        Returns
        -------
        TSVDataset : the dataset of target segment.
        """
        raise NotImplementedError

    def dataset_train(self):
        """Get the training segment of the dataset for the task.

        Returns
        -------
        tuple of str, TSVDataset : the segment name, and the dataset.
        """
        return 'train', self.get_dataset(segment='train')

    def dataset_dev(self):
        """Get the development (i.e. validation) segment of the dataset for this task.

        Returns
        -------
        tuple of (str, TSVDataset), or list of tuple : the segment name, and the dataset.
        """
        return 'dev', self.get_dataset(segment='dev')

    def dataset_test(self):
        """Get the test segment of the dataset for the task.

        Returns
        -------
        tuple of (str, TSVDataset), or list of tuple : the segment name, and the dataset.
        """
        return 'test', self.get_dataset(segment='test')

class ToySSTTask(AbstractGlueTask):
    """The Stanford Sentiment Treebank task on GLUE benchmark."""
    def __init__(self):
        is_pair = False
        class_labels = ['0', '1']
        self.metric = Accuracy()
        super(ToySSTTask, self).__init__(class_labels, self.metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for SST

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        dataset = nlp.data.GlueSST2(segment=segment)
        sampler, _ = get_split_samplers(dataset, split_ratio=0.01)
        return SampledDataset(dataset, sampler)

class AbstractCustomTask:
    """Abstract task class for custom datasets to be used with BERT-family models.

    Parameters
    ----------
    class_labels : list of str, or None
        Classification labels of the task.
        Set to None for regression tasks with continuous real values (TODO).
    metrics : list of EValMetric
        Evaluation metrics of the task.
    is_pair : bool
        Whether the task deals with sentence pairs or single sentences.
    label_alias : dict
        label alias dict, some different labels in dataset actually means
        the same. e.g.: {'contradictory':'contradiction'} means contradictory
        and contradiction label means the same in dataset, they will get
        the same class id.
    """
    def __init__(self, class_labels, metrics, is_pair, label_alias=None):
        self.class_labels = class_labels
        self.metrics = metrics
        self.is_pair = is_pair
        self.label_alias = label_alias

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for the task.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments.

        Returns
        -------
        Dataset : the dataset of target segment.
        """
        raise ValueError('This is an abstract class. For custom dataset support, '
                         'please use CustomTSVClassificationTask instead.')

    def dataset_train(self):
        """Get the training segment of the dataset for the task.

        Returns
        -------
        tuple of str, Dataset : the segment name, and the dataset.
        """
        return 'train', self.get_dataset(segment='train')

    def dataset_dev(self):
        """Get the development (i.e. validation) segment of the dataset for this task.

        Returns
        -------
        tuple of str, Dataset : the segment name, and the dataset.
        """
        return 'dev', self.get_dataset(segment='dev')

    def dataset_test(self):
        """Get the test segment of the dataset for the task.

        Returns
        -------
        tuple of str, Dataset : the segment name, and the dataset.
        """
        return 'test', self.get_dataset(segment='test')


class CustomTSVClassificationTask(AbstractCustomTask):
    """

    Parameters
    ----------
    filepath: str, path of the file
        Any valid string path is acceptable.
    sep : str, default ‘,’
        Delimiter to use.
    usecols : list-like or callable
        Return a subset of the columns.
    header : int, list of int, default 'infer', optional
        Row number(s) to use as the column names, and the start of the
        data. Default behavior is to infer the column names: if no names are passed the behavior is identical
        to header=0 and column names are inferred from the first line of the file,
        if column names are passed explicitly then the behavior is identical to header=None.
    names : array-like, optional
        List of column names to use.
    split_ratio: float, default 0.8, optional
        Split ratio of training data and HPO validation data. split_ratio of the data goes to training data,
        (1-split_ratio) of th data goes to HPO validation data.
    For other keyword arguments, please see
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html#pandas.read_csv`_ .
    """
    def __init__(self, *args, **kwargs):
        self._read(**kwargs)
        self.is_pair = False
        self._label_column_id = self._index_label_column(**kwargs)
        self.class_labels = list(set([sample[self._label_column_id] for sample in self.dataset]))
        self.metric = Accuracy()
        super(CustomTSVClassificationTask, self).__init__(self.class_labels, self.metric, self.is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for the task.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments.

        Returns
        -------
        The dataset of target segment.
        """
        if segment == 'train':
            return SampledDataset(self.dataset, self.train_sampler)
        elif segment == 'dev':
            return SampledDataset(self.dataset, self.dev_sampler)
        else:
            raise ValueError('Please specify either train or dev dataset.')

    def _read(self, **kwargs):
        low_memory = kwargs.get('low_memory', None)
        if low_memory is None:
            kwargs['low_memory'] = False
        split_ratio = kwargs.pop('split_ratio', 0.8)
        path = kwargs.pop('filepath', None)
        kwargs['filepath_or_buffer'] = path
        dataset_df = kwargs.pop('df', None)
        if path is not None:
            if 'usecols' not in kwargs:
                raise ValueError('`usecols` must be provided to support the '
                                 'identification of text and target/label columns.')
            dataset_df = pd.read_csv(**kwargs)
        elif dataset_df is not None:
            dataset_df = dataset_df.copy(deep=True)
        else:
            raise ValueError('A dataset constructed from either an external csv/tsv file by setting filepath=x,'
                             'or an in-memory python DataFrame by setting df=x is supported now.')
        dataset_df_lst = dataset_df.values.tolist()
        self.dataset = gluon.data.SimpleDataset(dataset_df_lst)
        self.train_sampler, self.dev_sampler = get_split_samplers(self.dataset, split_ratio=split_ratio)

    def _index_label_column(self, **kwargs):
        label_column_id = kwargs.get('label', None)
        if label_column_id is None:
            warnings.warn('By default we are using the last column as the label column, '
                          'if you wish to specify the label column by yourself or the column is not label column,'
                          'please specify label column using either the column name or index by using label=x.')
            return len(self.dataset[0]) - 1
        if isinstance(label_column_id, str):
            usecols = kwargs.get('usecols', None)
            if usecols is None:
                raise ValueError('Please specify the `usecols` when specifying the label column.')
            else:
                for i, col in enumerate(usecols):
                    if col == label_column_id:
                        return i
        elif isinstance(label_column_id, int):
            return label_column_id
        else:
            raise ValueError('Please specify label column using either the column name (str) or index (int).')

    def dataset_train(self):
        """Get the training segment of the dataset for the task.

        Returns
        -------
        tuple of str, Dataset : the segment name, and the dataset.
        """
        return 'train', self.get_dataset('train')

    def dataset_dev(self):
        """Get the dev segment of the dataset for the task.

        Returns
        -------
        tuple of str, Dataset : the segment name, and the dataset.
        """
        return 'dev', self.get_dataset('dev')

class MRPCTask(AbstractGlueTask):
    """The MRPC task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['0', '1']
        metric = CompositeEvalMetric()
        metric.add(F1())
        metric.add(Accuracy())
        super(MRPCTask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for MRPC.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueMRPC(segment=segment)

class QQPTask(AbstractGlueTask):
    """The Quora Question Pairs task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['0', '1']
        metric = CompositeEvalMetric()
        metric.add(F1())
        metric.add(Accuracy())
        super(QQPTask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for QQP.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueQQP(segment=segment)


class RTETask(AbstractGlueTask):
    """The Recognizing Textual Entailment task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['not_entailment', 'entailment']
        metric = Accuracy()
        super(RTETask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for RTE.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueRTE(segment=segment)

class QNLITask(AbstractGlueTask):
    """The SQuAD NLI task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['not_entailment', 'entailment']
        metric = Accuracy()
        super(QNLITask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for QNLI.

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueQNLI(segment=segment)

class STSBTask(AbstractGlueTask):
    """The Sentence Textual Similarity Benchmark task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = None
        metric = PearsonCorrelation()
        super(STSBTask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for STSB

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueSTSB(segment=segment)

class CoLATask(AbstractGlueTask):
    """The Warstdadt acceptability task on GLUE benchmark."""
    def __init__(self):
        is_pair = False
        class_labels = ['0', '1']
        metric = MCC(average='micro')
        super(CoLATask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for CoLA

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueCoLA(segment=segment)

class SSTTask(AbstractGlueTask):
    """The Stanford Sentiment Treebank task on GLUE benchmark."""
    def __init__(self):
        is_pair = False
        class_labels = ['0', '1']
        self.metric = Accuracy()
        super(SSTTask, self).__init__(class_labels, self.metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for SST

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'train', 'dev', 'test'.
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueSST2(segment=segment)

class WNLITask(AbstractGlueTask):
    """The Winograd NLI task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['0', '1']
        metric = Accuracy()
        super(WNLITask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for WNLI

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'dev', 'test', 'train'
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueWNLI(segment=segment)

class MNLITask(AbstractGlueTask):
    """The Multi-Genre Natural Language Inference task on GLUE benchmark."""
    def __init__(self):
        is_pair = True
        class_labels = ['neutral', 'entailment', 'contradiction']
        metric = Accuracy()
        super(MNLITask, self).__init__(class_labels, metric, is_pair)

    def get_dataset(self, segment='train'):
        """Get the corresponding dataset for MNLI

        Parameters
        ----------
        segment : str, default 'train'
            Dataset segments. Options are 'dev_matched', 'dev_mismatched', 'test_matched',
            'test_mismatched', 'train'
        """
        nlp = try_import_gluonnlp()
        return nlp.data.GlueMNLI(segment=segment)

    def dataset_dev(self):
        """Get the dev segment of the dataset for the task.

        Returns
        -------
        list of TSVDataset : the dataset of the dev segment.
        """
        return [('dev_matched', self.get_dataset(segment='dev_matched')),
                ('dev_mismatched', self.get_dataset(segment='dev_mismatched'))]

    def dataset_test(self):
        """Get the test segment of the dataset for the task.

        Returns
        -------
        list of TSVDataset : the dataset of the test segment.
        """
        return [('test_matched', self.get_dataset(segment='test_matched')),
                ('test_mismatched', self.get_dataset(segment='test_mismatched'))]

built_in_tasks = {
    'mrpc': MRPCTask,
    'qqp': QQPTask,
    'qnli': QNLITask,
    'rte': RTETask,
    'sts-b': STSBTask,
    'cola': CoLATask,
    'mnli': MNLITask,
    'wnli': WNLITask,
    'sst': SSTTask,
    'toysst': ToySSTTask,
}
